// -*- C++ -*-
// $RCSfile: boundarycond.C,v $
// $Revision: 1.29 $
// $Author: langer $
// $Date: 2007-01-29 00:10:29 $

/* This software was produced by NIST, an agency of the U.S. government,
 * and by statute is not subject to copyright in the United States.
 * Recipients of this software assume all responsibilities associated
 * with its operation, modification and maintenance. However, to
 * facilitate maintenance we ask that before distributing modified
 * versions of this software, you first contact the authors at
 * oof_manager@ctcms.nist.gov. 
 */

#include <oofconfig.h>

#include "common/trace.h"
#include "engine/boundarycond.h"
#include "engine/edge.h"
#include "engine/edgeset.h"
#include "engine/element.h"
#include "engine/equation.h"
#include "engine/field.h"
#include "engine/flux.h"
#include "engine/fluxnormal.h"
#include "engine/gausspoint.h"
#include "engine/nodalequation.h"
#include "engine/node.h"
#include "engine/sparselink.h"
#include "engine/csubproblem.h"
#include <iostream>
#include <stdlib.h>
#include <unistd.h>

class Material;
class Element;


// Functions to retrieve the mapped-to index that will be
// used if you call things directly.  This is used by a 
// FloatBC that doesn't know the index yet, to get it.
int FloatBCApp::getdofindex(CSubProblem *m, FuncNode *n, 
			    Field *field, int fcomp) {
  FuncNode &nref = *n;
  int doflistindex = (*field)(nref, fcomp)->dofindex();
  return m->freedofmap[doflistindex];
}

int FloatBCApp::geteqnindex(CSubProblem *m, FuncNode *n,
			    Equation *eqn, int eqcomp) {
  int ndqlistindex = m->map_forward_index(*eqn, *n, eqcomp);
  return m->indepeqnmap[ndqlistindex];
}


// This class is a bit of a special case, it's the C implementation
// of the "apply" function from the Python FloatBC class.  It does
// the leg-work of changing the mapping lists so that they're
// many-to-not-quite-so-many, aggregating DOF's and nodal equations
// correctly.

// On the very first call, the function needs to make a note
// of the mapped-to number it finds in the list, and on
// subsequent calls, set the others to that.  Then, by way
// of clean-up, it needs to ensure that the "target space"
// of the map is still contiguous in the integers [0..n], 
// possibly with a reduced n.  Note that some
// of the replaced entries might be smaller than the initial
// entry.

// "initprof" at this point is the value of the first evaluation 
// of the profile function.
void FloatBCApp::editmap(CSubProblem *m, double val, FuncNode *n, Field *field,
			 int fcomp, Equation *eqn, int eqcomp,
			 int newdofmap, int neweqnmap, double initprof) {

  FuncNode &nref = *n; 
  int doflistindex = (*field)(nref, fcomp)->dofindex();
  int ndqlistindex = m->map_forward_index(*eqn, *n, eqcomp);

  m->freedofmap[doflistindex] = newdofmap;
  m->indepeqnmap[ndqlistindex] = neweqnmap;

  // At this point, we want to accumulate the indexing information 
  // that will allow us to build the right "rhssub" matrix in "cleanlists."
  // We need to know which columns of the main stiffness matrix 
  // correspond to which components of "cvector", so we use 
  // yet another index-mapping integer array.
  
  profile_vector.push_back(val-initprof);
  rhssubtractor_index.push_back(doflistindex);


//   std::cerr << "Mapped equation list:" << std::endl;
//   for(int i=0; i<m->indepeqnmap.size(); i++) {
//     std::cerr << i << " : " << m->indepeqnmap[i] << std::endl;
//   }
//   std::cerr << "Mapped DOF list:" << std::endl;
//   for(int i=0; i<m->freedofmap.size(); i++) {
//     std::cerr << i << " : " << m->freedofmap[i] << std::endl;
//   }
}


void FloatBCApp::profile_rhs(CSubProblem *subproblem) {
  
  // Makes mods to the boundary-conditions' contribution to the rhs.
  // Uses the data in cvector (profile function offsets) and
  // rhssubtractor_index (column-number map).  For each row, you
  // need to subtract from the right-hand-side a quantity equal
  // to the stiffness matrix element times the difference in the
  // profile function from the initial value, i.e. the value at
  // the first-encountered node.  
  // This is done by making a map of the column-indices that were
  // encountered in "editmap", then looping over the entire 
  // stiffness matrix and doing the appropriate subtraction for
  // each element that comes from a mapped column.

  const SparseLinkMatrix<double> &stiffness = subproblem->stiffness_matrix();
  VECTOR_D &boundaryrhs(subproblem->boundary_rhs());
  std::vector<int> local_colmap(subproblem->stiffness_matrix().ncols(), -1);

  // Make the map of where to find the indices which were encountered.
  for(std::vector<int>::size_type i=0; i<local_colmap.size(); i++) {
    for(std::vector<int>::size_type j=0; j<rhssubtractor_index.size(); j++) {
      if (rhssubtractor_index[j]==int(i)) {
	local_colmap[i]=j;
	break;  // Only one hit per row.
      }
    }
  }
  
  // Loop over the stiffness matrix, pulling out the
  // values for which the rhs requires a contribution.
  for( CSparseLinkIterator<double> csi = stiffness.begin();
       csi!=stiffness.end(); ++csi ) {
    int k;
    if ((k = local_colmap[csi.col()])!=-1) {
      boundaryrhs[csi.row()] -= profile_vector[k]*(*csi);
    }
  }
}


//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//=\\=//

// Do the integration to apply a flux boundary conditon.
// The specified condition is in the form of a Python-callable
// object, which appears here as a PyObject*.

// Specified flux might vary in space, so there's a callback.
// What is actually applied is the normal component of the flux,
// which is a force density with the same dimensionality as the
// divergence of the flux. 
void NeumannBCApp::integrate(PyObject *wrapper, PyObject *pyfun, 
			     bool normal, double time) {
//   Trace("NeumannBCApp::integrate_");
  // Instance the iterator outside the "for" statement so that the
  // call to total_length is outside the loop.
  EdgeSetIterator bi = EdgeSetIterator(bdy);
  double edgeset_length = bi.total_length();

  for(; !bi.end(); ++bi) {
    
    BoundaryEdge* ed = bi.edge();
    const Element *el = ed->el;

    if(subproblem->contains(el)) {
      int order = el->shapefun_degree();
    
      order+=3; // Add a small offset to the shapefunction order.
      // Ultimately this should query the profile function for its
      // order somehow, but since the profile function is
      // user-specified, that's a bit tricky.
    
      double traversed_length = bi.traversed_length();
      double edge_length = ed->lab_length();
      for(EdgeGaussPoint egpt=ed->integrator(order); !egpt.end(); ++egpt) {
	double distance = traversed_length + edge_length*egpt.fraction();
	double edgeset_fraction = distance/edgeset_length;
	// Make the contribution to the RHS through the nodalequation.
	// Automatically calls the one with the correct Python
	// call-back wrapper.

	// Make the Python callback at the highest possible level,
	// outside as many loops as possible, because it's expensive.
	FluxNormal *flxnormal = flux->BCCallback(egpt.position(), 
						 time, 
						 egpt.normal(),
						 distance,
						 edgeset_fraction,
						 wrapper,
						 pyfun);
	// egpt.normal() is the normal at this point -- do the
	// rotation, if required.
	if (normal) 
	  flxnormal->transform(egpt.normal());
	flux->boundary_integral(subproblem, ed, egpt, flxnormal);
	delete flxnormal;
      }
    }
  }
}



void applyForceBC(CSubProblem *subproblem, Equation *equation, FuncNode *node,
		  int eqcomp, double val)
{
  // Asks the equation for the index corresponding to the current
  // component, then asks the mesh for that component of the
  // boundary_rhs vector, and subtracts val from it.
  int ndqlistindex = subproblem->map_forward_index(*equation, *node, eqcomp);
  subproblem->boundary_rhs()[ndqlistindex] -= val; 
}

// TODO LATER: PeriodicFlux boundary conditions.  The normal fluxes
// through two edges are constrained to be equal and opposite through
// the use of a Lagrange multiplier field.
